/* GtkZephyr - a GTK+ based front end to MIT's Zephyr Notification Service
 * Copyright (C) 2000-2001 Sean Dilda
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include <zephyr/zephyr.h>
#include <gtk/gtk.h>
#include <string.h>
#include <time.h>
#include "prefs.h"
#include "gtkzephyr.h"
#include "userdialog.h"
#include "chatdialog.h"
#include "zcalls.h"
#include "adduser.h"
#include "fileio.h"
#include "subscribe.h"
#include "yesno.h"

GHashTable *usertable, *insttable;
GtkWidget *mainwindow, *userclist, *instclist, *statusbar, *msgcounter;
GtkPaned *vpaned;
gint redraw_count;
gint newmsg_count;
guint msgcounter_context;

gint update_userlist();


gtkzephyr_user *gtkzephyr_get_user(gchar *userrealm) {
  DEBUG_INFO("Looking for %s..\n", userrealm);
  return (gtkzephyr_user *) g_hash_table_lookup(usertable, userrealm);
}

gtkzephyr_inst *gtkzephyr_get_inst(gchar *class, gchar *inst_str, gchar *recipient) {

  gchar *key = gtkzephyr_make_key(class, inst_str, recipient);
  gtkzephyr_inst *inst;
  inst = GTKZ_INST(g_hash_table_lookup(insttable, key));
  g_free(key);
  return inst;

}



void append_to_userclist(gpointer foo, gtkzephyr_user *user, gpointer bar) {

  gint rowcount;
  gchar *rowtext[2] = {NULL,NULL};

  rowcount = gtk_clist_append(GTK_CLIST(userclist), rowtext);
  gtk_clist_set_text(GTK_CLIST(userclist),rowcount,1,user->userid);
  if (GTKZ_SUBSCRIP(user)->unreadmsg) {
    gchar *unreadmsg_str = g_strdup_printf("%d", GTKZ_SUBSCRIP(user)->unreadmsg);
    gtk_clist_set_text(GTK_CLIST(userclist),rowcount,0, unreadmsg_str);
    g_free(unreadmsg_str);
  }
  gtk_clist_set_row_data(GTK_CLIST(userclist),rowcount,user);

}



void append_to_instclist(gtkzephyr_inst *inst) {

  gint rowcount;
  gchar *rowtext[2] = {NULL, NULL};

  rowcount = gtk_clist_append(GTK_CLIST(instclist), rowtext);
  gtk_clist_set_text(GTK_CLIST(instclist), rowcount, 1, inst->clistname);
  if (GTKZ_SUBSCRIP(inst)->unreadmsg) {
    gchar *unreadmsg_str = g_strdup_printf("%d", GTKZ_SUBSCRIP(inst)->unreadmsg);
    gtk_clist_set_text(GTK_CLIST(instclist), rowcount, 0, unreadmsg_str);
    g_free(unreadmsg_str);
  }
  gtk_clist_set_row_data(GTK_CLIST(instclist), rowcount, inst);

}



void append_to_clist(gtkzephyr_subscrip *subscrip) {

  if (subscrip->type == GTKZ_USER_TYPE) {
    append_to_userclist(NULL, GTKZ_USER(subscrip), NULL);
  } else if (subscrip->type == GTKZ_INST_TYPE) {
    append_to_instclist(GTKZ_INST(subscrip));
  } else
    g_assert_not_reached();

}



void gtkzephyr_add_unreadmsg(gtkzephyr_subscrip *subscrip) {

  GtkCList *clist = (subscrip->type == GTKZ_USER_TYPE ? GTK_CLIST(userclist) : GTK_CLIST(instclist));
  gint rownumber = gtk_clist_find_row_from_data(GTK_CLIST(clist), subscrip);
  gchar *unreadmsg_str = g_strdup_printf("%d", ++subscrip->unreadmsg);
  gchar *msgcounter_str = g_strdup_printf("%d Unread Msgs", ++newmsg_count);

  
  if (rownumber == -1) {
    append_to_clist(subscrip);
  } else {
    gtk_clist_set_text(GTK_CLIST(clist), rownumber, 0, unreadmsg_str);
  }
  g_free(unreadmsg_str);

  gtk_statusbar_push(GTK_STATUSBAR(msgcounter), msgcounter_context,
      msgcounter_str);
  g_free(msgcounter_str);

  if (redraw_count < prefstruct->max_redraw) {
    gtk_clist_sort(GTK_CLIST(clist));
    if (++redraw_count == prefstruct->max_redraw)
      gtk_clist_freeze(GTK_CLIST(clist));
  }
}



void gtkzephyr_remove_unreadmsg(gtkzephyr_subscrip *subscrip) {

  GtkCList *clist = (subscrip->type == GTKZ_USER_TYPE ? GTK_CLIST(userclist) : GTK_CLIST(instclist));
  gint rownumber = gtk_clist_find_row_from_data(GTK_CLIST(clist), subscrip);
  gchar *unreadmsg_str = g_strdup_printf("%d", --subscrip->unreadmsg);

  g_assert(subscrip->unreadmsg >= 0);
  g_assert(rownumber >= 0);
  if (subscrip->unreadmsg) {
    DEBUG_INFO("Setting text to %s\n", unreadmsg_str);
    gtk_clist_set_text(GTK_CLIST(clist), rownumber, 0,
        subscrip->unreadmsg ? unreadmsg_str : "");
  } else {
    DEBUG_INFO("Setting null text\n");
    if (subscrip->type == GTKZ_INST_TYPE || GTKZ_USER(subscrip)->status || prefstruct->show_disconnected) {
      DEBUG_INFO("Setting null text\n");
      gtk_clist_set_text(GTK_CLIST(clist), rownumber, 0, "");
    } else {
      DEBUG_INFO("Removing user from clist\n");
      gtk_clist_remove(GTK_CLIST(clist), rownumber);
    }
  }
  g_free(unreadmsg_str);

  newmsg_count--;
  g_assert(newmsg_count >= 0);
  gtk_statusbar_pop(GTK_STATUSBAR(msgcounter), msgcounter_context);


  if (redraw_count < prefstruct->max_redraw) {
    gtk_clist_sort(GTK_CLIST(clist));
    if (++redraw_count == prefstruct->max_redraw)
      gtk_clist_freeze(GTK_CLIST(clist));
  }
}



void gtkzephyr_handle_user_login(gtkzephyr_user *user) {

  int rownumber = gtk_clist_find_row_from_data(GTK_CLIST(userclist), user);
  /* rownumber will only be -1 (not in clist) if we don't show disconnected
   * users and they haven't already been added by sending us message */
  if (rownumber == -1) {
    DEBUG_INFO("Adding %s to userclist..\n", user->userid);
    append_to_userclist(NULL, user, NULL);
  }
  user->status = 1;

}



void gtkzephyr_handle_user_logout(gtkzephyr_user *user) {

  gtkzephyr_subscrip *subscrip = (gtkzephyr_subscrip *) user;

  if (!subscrip->unreadmsg) {
    if (!prefstruct->show_disconnected) {
      DEBUG_INFO("Removing %s from userclist..\n", user->userid);
      gtk_clist_remove(GTK_CLIST(userclist), gtk_clist_find_row_from_data(
            GTK_CLIST(userclist), user));
    }
  }

  user->status = 0;

}
    



void gtkzephyr_add_user(gchar *userid) {

  gtkzephyr_file_add_user(userid);
  gtkzephyr_make_user(userid);
  update_userlist();

}



gint userclist_comparefunc(GtkCList *clist, gconstpointer a, gconstpointer b) {

  gtkzephyr_user *usera = (gtkzephyr_user*)((GtkCListRow *)a)->data;
  gtkzephyr_user *userb = (gtkzephyr_user*)((GtkCListRow *)b)->data;

  if (GTKZ_SUBSCRIP(usera)->unreadmsg == 0 && GTKZ_SUBSCRIP(userb)->unreadmsg == 0) {
    if (usera->status < userb->status) {
      return 1;
    } else {
      return -1;
    }
  } else {
    if (GTKZ_SUBSCRIP(usera)->unreadmsg < GTKZ_SUBSCRIP(userb)->unreadmsg) {
      return 1;
    } else {
      return -1;
    }
  }
}



gint instclist_comparefunc(GtkCList *clist, gconstpointer a, gconstpointer b) {

  gtkzephyr_subscrip *subscripa = GTKZ_SUBSCRIP(((GtkCListRow *)a)->data);
  gtkzephyr_subscrip *subscripb = GTKZ_SUBSCRIP(((GtkCListRow *)b)->data);

  gint rv = subscripa->unreadmsg < subscripb->unreadmsg;

  return rv ? 1 : -1;

}



void clear_subscrip(gtkzephyr_subscrip *subscrip) {

  gint i;

  for (i = 0; i < subscrip->unreadmsg; i++) {
    gtk_statusbar_pop(GTK_STATUSBAR(msgcounter), msgcounter_context);
    newmsg_count--;
  }
  g_list_foreach(subscrip->readmsgs, (GFunc) ZFreeNotice, NULL);
  g_list_foreach(subscrip->newmsgs, (GFunc) ZFreeNotice, NULL);
  g_list_free(subscrip->readmsgs);
  g_list_free(subscrip->newmsgs);
  if (subscrip->menu)
    gtk_widget_destroy(subscrip->menu);
  if (subscrip->userdialog)
    gtk_widget_destroy(subscrip->userdialog);
  if(subscrip->chatwindow)
    gtk_widget_destroy(subscrip->chatwindow);

}



void gtkzephyr_remove_user(gtkzephyr_user *user) {

  g_hash_table_remove(usertable, user->userrealm);
  gtk_clist_remove(GTK_CLIST(userclist),
      gtk_clist_find_row_from_data(GTK_CLIST(userclist), user));
  gtkzephyr_file_remove_user(user->userid);
  clear_subscrip(GTKZ_SUBSCRIP(user));
  g_free(user->userid);
  g_free(user->userrealm);
  g_free(user);

}



void remove_user_cb(GtkWidget *widget, gtkzephyr_user *user) {

  gchar *msg = g_strdup_printf("Are you sure you want to remove %s?",
      user->userid);

  gtkzephyr_yesno(msg, (GtkFunction) gtkzephyr_remove_user, user);

  g_free(msg);

}



void gtkzephyr_remove_inst(gtkzephyr_inst *inst) {

  g_hash_table_remove(insttable, inst->key);
  gtk_clist_remove(GTK_CLIST(instclist),
      gtk_clist_find_row_from_data(GTK_CLIST(instclist), inst));
  gtkzephyr_file_remove_subscrip(inst->class, inst->instance, inst->recipient);
  clear_subscrip(GTKZ_SUBSCRIP(inst));
  g_free(inst->class);
  g_free(inst->instance);
  g_free(inst->recipient);
  if (inst->key != inst->clistname)
    g_free(inst->clistname);
  g_free(inst->key);
  g_free(inst);

}



void remove_inst_cb(GtkWidget *widget, gtkzephyr_inst *inst) {

  gchar *msg = g_strdup_printf("Are you sure you want to remove the instance %s?", inst->clistname);

  gtkzephyr_yesno(msg, (GtkFunction) gtkzephyr_remove_inst, inst);

  g_free(msg);

}



#define unsensitive_menu_item(x)  menu_item = gtk_menu_item_new_with_label(x); \
                                  gtk_menu_append(GTK_MENU(menu), menu_item); \
                                  gtk_widget_set_sensitive(menu_item, FALSE); \
                                  gtk_widget_show(menu_item);
#define menu_item(label, cb) menu_item = gtk_menu_item_new_with_label(label); \
                             gtk_menu_append(GTK_MENU(menu), menu_item); \
                             gtk_signal_connect(GTK_OBJECT(menu_item), "activate", \
                             GTK_SIGNAL_FUNC(cb), NULL); \
                             gtk_widget_show(menu_item);

#define menu_item2(label, cb, pointer) menu_item = gtk_menu_item_new_with_label(label); \
                                       gtk_menu_append(GTK_MENU(menu), menu_item); \
                                       gtk_signal_connect(GTK_OBJECT(menu_item), "activate", GTK_SIGNAL_FUNC(cb), pointer); \
                                       gtk_widget_show(menu_item);



void build_menus(GtkMenuBar *menubar) {

  GtkWidget *menu, *menu_item;

  menu = gtk_menu_new();
  unsensitive_menu_item("Send Zephyr");
  menu_item("Add User", gtkzephyr_add_user_dialog);
  unsensitive_menu_item("Preferences");
  menu_item("Exit", gtk_main_quit);

  menu_item = gtk_menu_item_new_with_label("Program");
  gtk_menu_item_set_submenu(GTK_MENU_ITEM(menu_item), menu);
  gtk_menu_bar_append(GTK_MENU_BAR(menubar), menu_item);
  gtk_widget_show(menu_item);

  menu = gtk_menu_new();
  menu_item("Flush locations", gtkzephyr_flush_loc);
  menu_item("Hide", gtkzephyr_hide);
  menu_item("Unhide", gtkzephyr_unhide);
  menu_item("Subscribe", gtkzephyr_subscribe_dialog);
  unsensitive_menu_item("Unubscribe");

  menu_item = gtk_menu_item_new_with_label("Zctl");
  gtk_menu_item_set_submenu(GTK_MENU_ITEM(menu_item), menu);
  gtk_menu_bar_append(GTK_MENU_BAR(menubar), menu_item);
  gtk_widget_show(menu_item);
  
}



GtkWidget *make_user_menu(gtkzephyr_user *user) {

  GtkWidget *menu, *menu_item;

  menu = gtk_menu_new();
  menu_item2("Read Message", gtkzephyr_open_read, user);
  menu_item2("Send Message", gtkzephyr_open_write, user);
  menu_item2("Open Chat", gtkzephyr_open_chat_window, user);
  unsensitive_menu_item("Get Info");
  menu_item2("Remove User", remove_user_cb, user);

  return menu;
}



GtkWidget *make_inst_menu(gtkzephyr_inst *inst) {

  GtkWidget *menu, *menu_item;

  menu = gtk_menu_new();
  menu_item2("Read Message", gtkzephyr_open_read, inst);
  menu_item2("Send Message", gtkzephyr_open_write, inst);
  menu_item2("Open Chat", gtkzephyr_open_chat_window, inst);
  menu_item2("Remove Instance", remove_inst_cb, inst);

  return menu;

}



void row_selected(GtkWidget *clist, GdkEventButton *event, gpointer data) {
  gint row, column;
  gtkzephyr_subscrip *subscrip;

  if (gtk_clist_get_selection_info(GTK_CLIST(clist), event->x, event->y, &row, &column)) {
    gtk_clist_select_row(GTK_CLIST(clist), row, column);
    subscrip = gtk_clist_get_row_data(GTK_CLIST(clist), GPOINTER_TO_INT(row));
    if (event->type == GDK_BUTTON_PRESS && event->button == 3) {
      if (subscrip->menu == NULL)
        subscrip->menu = (subscrip->type == GTKZ_USER_TYPE) ? make_user_menu(
            GTKZ_USER(subscrip)) : make_inst_menu(GTKZ_INST(subscrip));
      gtk_menu_popup(GTK_MENU(subscrip->menu), NULL, NULL, NULL, NULL,
          event->button, event->time);
    } else if (event->type == GDK_2BUTTON_PRESS && event->button == 1) {
      if (prefstruct->userd_on_double_cl)
        gtkzephyr_user_dialog(subscrip);
      else
        gtkzephyr_chat_window(subscrip);
    } else if (event->type == GDK_2BUTTON_PRESS && event->button == 2) {
      if (prefstruct->userd_on_double_cl)
        gtkzephyr_chat_window(subscrip);
      else
        gtkzephyr_user_dialog(subscrip);
    }
  }
}



void build_main_window() {

  GtkWidget *vbox, *menubar, *scrolledwindow;

  mainwindow = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_window_set_default_size(GTK_WINDOW(mainwindow), 125, 300);
  gtk_signal_connect(GTK_OBJECT(mainwindow), "delete_event",
      (GtkSignalFunc) gtk_main_quit, NULL);
  gtk_window_set_title(GTK_WINDOW(mainwindow), "GtkZephyr");
  gtk_container_set_border_width(GTK_CONTAINER(mainwindow), 5);
  
  vbox = gtk_vbox_new(FALSE, 0);
  gtk_container_add(GTK_CONTAINER(mainwindow), vbox);
  gtk_widget_show(vbox);

  menubar = gtk_menu_bar_new();
  gtk_box_pack_start(GTK_BOX(vbox), menubar, FALSE, FALSE, 3);
  gtk_widget_show(menubar);

  build_menus(GTK_MENU_BAR(menubar));

  vpaned = GTK_PANED(gtk_vpaned_new());
  userclist = gtk_clist_new(2);
  gtk_clist_set_column_width(GTK_CLIST(userclist),0,8);
  gtk_clist_set_column_title(GTK_CLIST(userclist),1, "Userid");
  gtk_clist_column_titles_show(GTK_CLIST(userclist));
  gtk_clist_set_compare_func(GTK_CLIST(userclist), userclist_comparefunc);
  gtk_signal_connect(GTK_OBJECT(userclist), "button_press_event",
                     GTK_SIGNAL_FUNC(row_selected), NULL);

  scrolledwindow = gtk_scrolled_window_new(GTK_CLIST(userclist)->hadjustment,
                                           GTK_CLIST(userclist)->vadjustment);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolledwindow),
                                 GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
  gtk_widget_show(scrolledwindow);
  gtk_container_add(GTK_CONTAINER(scrolledwindow), userclist);
  gtk_widget_show(userclist);
  gtk_paned_pack1(GTK_PANED(vpaned), scrolledwindow, TRUE, FALSE);

  instclist = gtk_clist_new(2);
  gtk_clist_set_column_width(GTK_CLIST(instclist),0,8);
  gtk_clist_set_column_title(GTK_CLIST(instclist),1, "Instance");
  gtk_clist_column_titles_show(GTK_CLIST(instclist));
  gtk_clist_set_compare_func(GTK_CLIST(instclist), instclist_comparefunc);
  gtk_signal_connect(GTK_OBJECT(instclist), "button_press_event",
                     GTK_SIGNAL_FUNC(row_selected), NULL);

  scrolledwindow = gtk_scrolled_window_new(GTK_CLIST(instclist)->hadjustment,
                                           GTK_CLIST(instclist)->vadjustment);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolledwindow),
                                 GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
  gtk_widget_show(scrolledwindow);
  gtk_container_add(GTK_CONTAINER(scrolledwindow), instclist);
  gtk_widget_show(instclist);
  gtk_paned_pack2(GTK_PANED(vpaned), scrolledwindow, FALSE, TRUE);
  gtk_widget_show(GTK_WIDGET(vpaned));
  gtk_box_pack_start(GTK_BOX(vbox), GTK_WIDGET(vpaned), TRUE, TRUE, 3);

  msgcounter = gtk_statusbar_new(); 
  gtk_box_pack_start(GTK_BOX(vbox), msgcounter, FALSE, FALSE, 3);
  msgcounter_context = gtk_statusbar_get_context_id(GTK_STATUSBAR(msgcounter),
      "Message Counter");
  gtk_statusbar_push(GTK_STATUSBAR(msgcounter), msgcounter_context,
      "No New Msgs");
  gtk_widget_show(msgcounter);

  statusbar = gtk_statusbar_new();
  gtk_box_pack_start(GTK_BOX(vbox), statusbar, FALSE, FALSE, 0);
  gtk_widget_show(statusbar);

}



gtkzephyr_user *gtkzephyr_make_user(gchar *userid) {

  gtkzephyr_user *user = g_new(gtkzephyr_user,1);
  GTKZ_SUBSCRIP(user)->type = GTKZ_USER_TYPE;
  user->userid = g_strdup(userid);
  if (!strchr(userid, '@')) {
    user->userrealm = g_strdup_printf("%s@%s", userid, ZGetRealm());
  } else {
    user->userrealm = g_strdup(userid);
  }
  DEBUG_INFO("userrealm: %s\n", user->userrealm);
  GTKZ_SUBSCRIP(user)->newmsgs = NULL;
  GTKZ_SUBSCRIP(user)->readmsgs = NULL;
  GTKZ_SUBSCRIP(user)->unreadmsg = 0;
  user->status = 0;
  GTKZ_SUBSCRIP(user)->menu = NULL;
  GTKZ_SUBSCRIP(user)->userdialog = NULL;
  GTKZ_SUBSCRIP(user)->notebook = NULL;
  GTKZ_SUBSCRIP(user)->newmsgclist = NULL;
  GTKZ_SUBSCRIP(user)->newmsgtext = NULL;
  GTKZ_SUBSCRIP(user)->dialogsendtext = NULL;
  GTKZ_SUBSCRIP(user)->dialogsendbutton = NULL;
  GTKZ_SUBSCRIP(user)->chatwindow = NULL;
  GTKZ_SUBSCRIP(user)->chattext = NULL;
  GTKZ_SUBSCRIP(user)->chatsendentry = NULL;
  GTKZ_SUBSCRIP(user)->chatsendbutton = NULL;
  g_hash_table_insert(usertable, user->userrealm, user);

  gtkzephyr_sub_login_class(user);

  return user;
}



gtkzephyr_inst *gtkzephyr_make_inst(gchar *class, gchar *instance, gchar *recipient) {

  gtkzephyr_inst *inst;
  gchar *key;

  key = gtkzephyr_make_key(class, instance, recipient);
  inst = g_hash_table_lookup(insttable, key);
  if (inst != NULL) {
    g_free(key);
    return inst;
  }

   
  inst = g_new(gtkzephyr_inst, 1);
  GTKZ_SUBSCRIP(inst)->type = GTKZ_INST_TYPE;
  GTKZ_SUBSCRIP(inst)->newmsgs = NULL;
  GTKZ_SUBSCRIP(inst)->readmsgs = NULL;
  GTKZ_SUBSCRIP(inst)->unreadmsg = 0;
  GTKZ_SUBSCRIP(inst)->menu = NULL;
  GTKZ_SUBSCRIP(inst)->userdialog = NULL;
  GTKZ_SUBSCRIP(inst)->notebook = NULL;
  GTKZ_SUBSCRIP(inst)->newmsgclist = NULL;
  GTKZ_SUBSCRIP(inst)->newmsgtext = NULL;
  GTKZ_SUBSCRIP(inst)->dialogsendtext = NULL;
  GTKZ_SUBSCRIP(inst)->dialogsendbutton = NULL;
  GTKZ_SUBSCRIP(inst)->chatwindow = NULL;
  GTKZ_SUBSCRIP(inst)->chattext = NULL;
  GTKZ_SUBSCRIP(inst)->chatsendentry = NULL;
  GTKZ_SUBSCRIP(inst)->chatsendbutton = NULL;
  inst->class = g_strdup(class);
  inst->instance = g_strdup(instance);
  inst->recipient = g_strdup(recipient);
  inst->key = key;
  inst->clistname = key;

  g_hash_table_insert(insttable, inst->key, inst);
  append_to_instclist(inst);

  return inst;

}



gchar *gtkzephyr_make_key(gchar *class, gchar *instance, gchar *recipient) {

  gchar *key;

  key = (strcmp(recipient, "*") && strcmp(recipient, "")) ?
    g_strdup_printf("%s,%s,%s", class, instance, recipient) :
    g_strdup_printf("%s,%s", class, instance);
  g_strup(key);

  return key;

}



void sub_to_insts(gpointer foo, gtkzephyr_inst *inst, gpointer bar) {

  gtkzephyr_sub(inst->class, inst->instance, inst->recipient);

}



gint update_userlist() {

  gtkzephyr_update_userlist(usertable);
  
  return TRUE;
}



gint redraw_clist(gpointer foo) {

  g_assert(redraw_count <= prefstruct->max_redraw);
  if (redraw_count == prefstruct->max_redraw) {
    /* gtk_clist_sort() will also thaw for us */
    gtk_clist_sort(GTK_CLIST(userclist));
  }

  redraw_count = 0;

  return TRUE;
}



gboolean hide_instclist(gpointer foo) {

  gtk_paned_set_position(vpaned, vpaned->max_position);

  return FALSE;

}



int main (int argc, char *argv[]) {

  gchar *key;
  gtkzephyr_inst *help_inst;

  gtk_init(&argc, &argv);

  build_main_window();
  gtkzephyr_connect();  /* should connect, subscribe, etc */
  gtkzephyr_load_prefs();
  usertable = g_hash_table_new(g_str_hash, g_str_equal);
  insttable = g_hash_table_new(g_str_hash, g_str_equal);
  gtkzephyr_process_user_file();
  gtkzephyr_process_subscrip_file();
  gtkzephyr_update_userlist(usertable);
  if (prefstruct->show_disconnected)
    g_hash_table_foreach(usertable, (GHFunc) append_to_userclist, NULL);
  g_hash_table_foreach(insttable, (GHFunc) sub_to_insts, NULL);
  if (prefstruct->help_instance && prefstruct->help_class) {
    key = gtkzephyr_make_key(prefstruct->help_class, prefstruct->help_instance,
	"");
    if (!(help_inst = g_hash_table_lookup(insttable, key))) {
      help_inst = gtkzephyr_make_inst(prefstruct->help_class,
          prefstruct->help_instance, "*");
      // inst->clistname = g_strdup("help");
    } 
    g_free(key);
    gtk_clist_set_text(GTK_CLIST(instclist),
        gtk_clist_find_row_from_data(GTK_CLIST(instclist), help_inst), 1, "help");
  }
  // g_hash_table_foreach(insttable, (GHFunc) append_to_instclist, NULL);
  gtkzephyr_flush_sub_list();
  gtkzephyr_flush_unsub_list();
  if (!GTK_CLIST(instclist)->rows) {
    gtk_timeout_add(1, (GtkFunction) hide_instclist, NULL);
  }
  gtk_widget_show(mainwindow);
  redraw_count = 0;
  newmsg_count = 0;

  gtk_timeout_add(prefstruct->check_messages_timeout,
      (GtkFunction)gtkzephyr_checknewnotices, GTK_CLIST(userclist));

  gtk_timeout_add(prefstruct->update_ul_timeout, (GtkFunction) update_userlist,
      NULL);

  
  gtk_timeout_add(prefstruct->redraw_ul_timeout, (GtkFunction) redraw_clist,
      NULL);

  gtk_main();

  // User can't change anything, so why save it.. later on make it so it
  // save only when user changes something, and immediately when they change
  // it
  // gtkzephyr_save_prefs();
  gtkzephyr_disconnect();

  return(0);
}
